#include <iostream>
#include <ctime>
#include <cmath>
#include <iomanip>
using namespace std;

#include "Timer.h"
#include "Progressbar.h"

double calculateExpectation(double *list, int size);
double calculateStdDev(double expectation, double *list, int size);
void zInterval(double expectation, double stdDev, int size); // Prints a ZInterval with C-level = 0.95
void zTest(double postExp, double preExp, double postStdDev, double preStdDev, int size);
double listSum(double *list,int size);

int main() 
{
  const int nIterations = 65535;
  
  double *preData, *postData;
  preData = new double[nIterations];
  postData = new double[nIterations];

  Timer timer;
  Progressbar progressbar(0,nIterations);
  int post = 0, pre = 0;

  cout << "++i vs. i++"
       << "\n - the statistical approach\n"
       << "Incrementing ";
  progressbar.print(true);

  for(int i = 0; i < nIterations; i++) {
    timer.start();
    post++;
    timer.stop();

    postData[i] = timer.getElapsed();

    timer.start();
    ++pre;
    timer.stop();

    preData[i] = timer.getElapsed();

    progressbar.increment();
    progressbar.print();
  }

  progressbar.clearBar(true);
  cout << "\b...done" << endl;

  double preExp = calculateExpectation(preData,nIterations), 
         preStdDev = calculateStdDev(preExp,preData,nIterations),
         postExp = calculateExpectation(postData,nIterations), 
         postStdDev = calculateStdDev(postExp,postData,nIterations);


  cout << setprecision(3)
       << "Post++\n"
       << " X: " << postExp 
       << "\n s: " << postStdDev 
       << "\n ZInterval: ";
  zInterval(postExp,postStdDev,nIterations);

  cout << "\n\n++Pre\n"
       << " X: " << preExp 
       << "\n s: " << preStdDev 
       << "\n ZInterval: ";
  zInterval(preExp,preStdDev,nIterations);
  
  cout << "\n\n";
  zTest(postExp,preExp,postStdDev,preStdDev,nIterations);
  cout << "\n\n";

  cin.get();

  delete [] preData;
  delete [] postData;

  return 0;
}

double calculateExpectation(double *list, int size)
{
  return listSum(list,size) / size;
}

double calculateStdDev(double expectation, double *list, int size)
{
  double *calc = new double[size], stddev;

  for(int i = 0; i < size; i++) {
    calc[i] = pow(list[i] - expectation,2.0);
  }

  stddev = sqrt(listSum(calc,size) / (size - 1));

  delete [] calc;

  return stddev;
}

double listSum(double *list, int size)
{
  double sum = 0;
  for(int i = 0; i < size; i++) {
    sum += list[i];
  }
  return sum;  
}

void zInterval(double expectation, double stdDev, int size)
{
  const double zval = 1.96;
  double range = zval * (stdDev / sqrt(static_cast<double>(size)));
  cout << setprecision(3) << "[" << expectation - range << "," << expectation + range << "]";
}

void zTest(double postExp, double preExp, double postStdDev, double preStdDev, int size)
{
  const double zval = 1.96;
  double Z = abs(postExp - preExp),
         C = zval * sqrt((pow(postStdDev,2.0) + pow(preStdDev,2.0)) / size);

  cout << "ZTest result: ";
  if (Z > C) {
    cout << "++i == i++" << endl;
  } else {
    cout << "++i != i++" << endl;
  }
}